import base64
import contextlib
import random
import hashlib
import logging
import os
import secrets
import time
import urllib
import webbrowser
from datetime import date, datetime
from enum import Enum
from functools import wraps
from http.server import BaseHTTPRequestHandler, HTTPServer
from importlib.metadata import version
from io import BytesIO
from typing import Any, Generator, get_type_hints, List, Literal
from urllib.parse import parse_qs, urlparse

import httpx
import pandas as pd

from .constants import DEFAULT_URL
from .schemas import (
    ActivityDetails, ActivitySummary, Metric, Sport, TraceDetails, UserInfoResponse, UserSummary
)
from .utils import decode_jwt_body, make_dataframe_streamlit_compatible


AUTH_SUCCESSFUL_RESPONSE = """<!DOCTYPE html>
<html>
<head>
    <style>
        body { max-width: 600px; margin: 40px auto; text-align: center; }
        h1 { color: #2C3E50; font-size: 24px; }
        p { color: #34495E; font-size: 18px; }
    </style>
</head>
<body>
    <img src="https://sweatstack.no/images/sweat-stack-python-client.png" alt="SweatStack Logo" style="width: 200px; margin: 20px auto; display: block;">
    <h1>Successfully authenticated with SweatStack!</h1>
    <p>You have successfully authenticated using the SweatStack Python client library. You can now close this window and return to your Python environment.</p>
</body>
</html>"""
OAUTH2_CLIENT_ID = "5382f68b0d254378"


try:
    __version__ = version("sweatstack")
except ImportError:
    __version__ = "unknown"


class OAuth2Mixin:
    def login(self):
        """Initiates the OAuth2 login flow for SweatStack authentication.

        This method starts a local HTTP server to receive the OAuth2 callback,
        opens a browser window for the user to authenticate with SweatStack,
        and exchanges the authorization code for an access token.

        The method uses PKCE (Proof Key for Code Exchange) for enhanced security
        during the OAuth2 authorization code flow.

        Returns:
            None

        Raises:
            Exception: If the authentication process times out or fails.

        Note:
            This method requires a working internet connection and the ability
            to open a browser window. It will also temporarily open a local HTTP
            server on a random port between 8000-9000.
        """
        class AuthHandler(BaseHTTPRequestHandler):
            def log_message(self, format, *args):
                # This override disables logging.
                pass

            def do_GET(self):
                query = urlparse(self.path).query
                params = parse_qs(query)
                
                self.server.code = params.get("code", [None])[0]
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(AUTH_SUCCESSFUL_RESPONSE.encode())
                self.server.server_close()

        code_verifier = secrets.token_urlsafe(32)
        code_challenge = hashlib.sha256(code_verifier.encode("ascii")).digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge).rstrip(b"=").decode("ascii")

        while True:
            port = random.randint(8000, 9000)
            try:
                server = HTTPServer(("localhost", port), AuthHandler)
                break
            except OSError:
                continue

        redirect_uri = f"http://localhost:{port}"
        params = {
            "client_id": OAUTH2_CLIENT_ID,
            "redirect_uri": redirect_uri,
            "code_challenge": code_challenge,
            "scope": "data:read profile",
            "prompt": "none",
        }
        base_url = self.url
        path = "/oauth/authorize"
        authorization_url = urllib.parse.urljoin(base_url, path + "?" + urllib.parse.urlencode(params))
        webbrowser.open(authorization_url)

        print(f"Waiting for authorization... (listening on port {port})")
        print(f"If not redirected, open the following URL in your browser: {authorization_url}")
        print("")

        server.timeout = 30
        try:
            server.handle_request()
        except TimeoutError:
            raise Exception("SweatStack Python login timed out after 30 seconds. Please try again.")

        if hasattr(server, "code"):
            token_data = {
                "grant_type": "authorization_code",
                "client_id": OAUTH2_CLIENT_ID,
                "code": server.code,
                "code_verifier": code_verifier,
            }
            response = httpx.post(
                f"{self.url}/api/v1/oauth/token",
                data=token_data,
            )
            try:
                self._raise_for_status(response)
            except httpx.HTTPStatusError as e:
                raise Exception(f"SweatStack Python login failed. Please try again.") from e
            token_response = response.json()

            self.jwt = token_response.get("access_token")
            self.api_key = self.jwt
            self.refresh_token = token_response.get("refresh_token")
            print(f"SweatStack Python login successful.")
        else:
            raise Exception("SweatStack Python login failed. Please try again.")


class DelegationMixin:
    def _validate_user(self, user: str | UserSummary):
        if isinstance(user, UserSummary):
            return user.id
        else:
            return user

    def _get_delegated_token(self, user: str | UserSummary):
        user_id = self._validate_user(user)
        with self._http_client() as client:
            response = client.post(
                "/api/v1/oauth/delegated-token",
                json={"sub": user_id},
            )
            self._raise_for_status(response)

        return response.json()

    def switch_user(self, user: str | UserSummary):
        """Switches the client to operate on behalf of another user.

        This method changes the current client's authentication to act on behalf of the specified user.
        The client will use a delegated token for all subsequent API calls.

        Args:
            user: Either a UserSummary object or a string user ID representing the user to switch to.

        Returns:
            None

        Raises:
            HTTPStatusError: If the delegation request fails.
        """
        token_response = self._get_delegated_token(user)
        self.api_key = token_response["access_token"]
        self.refresh_token = token_response["refresh_token"]

    def _get_principal_token(self):
        with self._http_client() as client:
            response = client.get(
                "/api/v1/oauth/principal-token",
            )
            self._raise_for_status(response)
        return response.json()

    def switch_back(self):
        """Switches the client back to the principal user.

        This method reverts the client's authentication from a delegated user back to the principal user.
        The client will use the principal token for all subsequent API calls.

        Returns:
            None

        Raises:
            HTTPStatusError: If the principal token request fails.
        """

        token_response = self._get_principal_token()
        self.api_key = token_response["access_token"]
        self.refresh_token = token_response["refresh_token"]

    def delegated_client(self, user: str | UserSummary):
        """Creates a new client instance that operates on behalf of another user.

        This method creates a new client instance with delegated authentication for the specified user.
        Unlike `switch_user`, this method does not modify the current client but returns a new one.

        Args:
            user: Either a UserSummary object or a string user ID representing the user to delegate to.

        Returns:
            Client: A new client instance authenticated as the delegated user.

        Raises:
            HTTPStatusError: If the delegation request fails.
        """
        token_response = self._get_delegated_token(user)
        return self.__class__(
            api_key=token_response["access_token"],
            refresh_token=token_response["refresh_token"],
            url=self.url,
            streamlit_compatible=self.streamlit_compatible,
        )

    def principal_client(self):
        """Creates a new client instance that operates as the principal user.

        This method creates a new client instance with authentication for the principal user.
        Unlike `switch_back`, this method does not modify the current client but returns a new one.

        Returns:
            Client: A new client instance authenticated as the principal user.

        Raises:
            HTTPStatusError: If the principal token request fails.
        """
        token_response = self._get_principal_token()
        return self.__class__(
            api_key=token_response["access_token"],
            refresh_token=token_response["refresh_token"],
            url=self.url,
            streamlit_compatible=self.streamlit_compatible,
        )


class Client(OAuth2Mixin, DelegationMixin):
    def __init__(
        self,
        api_key: str | None = None,
        refresh_token: str | None = None,
        url: str | None = None,
        streamlit_compatible: bool = False,
    ):
        self.api_key = api_key
        self.refresh_token = refresh_token
        self.url = url
        self.streamlit_compatible = streamlit_compatible

    def _do_token_refresh(self, tz_offset: int) -> str:
        with self._http_client() as client:
            response = client.post(
                "/api/v1/oauth/token",
                json={
                    "grant_type": "refresh_token",
                    "refresh_token": self.refresh_token,
                    "tz_offset": tz_offset,
                },
            )

            self._raise_for_status(response)
            return response.json()["access_token"]

    def _check_token_expiry(self, token: str) -> str:
        try:
            body = decode_jwt_body(token)
            # Margin in seconds to account for time to token validation of the next request
            TOKEN_EXPIRY_MARGIN = 5
            if body["exp"] - TOKEN_EXPIRY_MARGIN < time.time():
                # Token is (almost) expired, refresh it
                token = self._do_token_refresh(body["tz_offset"])
                self._api_key = token
        except Exception:
            # If token can't be decoded, just return as-is
            # @TODO: This probably should be handled differently
            pass

        return token

    @property
    def api_key(self) -> str:
        if self._api_key is not None:
            value = self._api_key
        else:
            value = os.getenv("SWEATSTACK_API_KEY")

        if value is None:
            # A non-authenticated client is a potentially valid use-case.
            return None

        return self._check_token_expiry(value)

    @api_key.setter
    def api_key(self, value: str):
        self._api_key = value
    
    @property
    def refresh_token(self) -> str:
        if self._refresh_token is not None:
            return self._refresh_token
        else:
            return os.getenv("SWEATSTACK_REFRESH_TOKEN")

    @refresh_token.setter
    def refresh_token(self, value: str):
        self._refresh_token = value

    @property
    def url(self) -> str:
        """
        This determines which SweatStack URL to use, allowing the use of a non-default instance.
        This is useful for example during local development.
        Please note that changing the url probably requires changing the `OAUTH2_CLIENT_ID` as well.
        """
        if self._url is not None:
            return self._url
        
        if env_url := os.getenv("SWEATSTACK_URL"):
            return env_url
            
        return DEFAULT_URL
    
    @url.setter
    def url(self, value: str):
        self._url = value
    
    @contextlib.contextmanager
    def _http_client(self):
        """
        Creates an httpx client with the base URL and authentication headers pre-configured.
        """
        headers = {
            "User-Agent": f"python-sweatstack/{__version__}",
        }
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        
        with httpx.Client(base_url=self.url, headers=headers) as client:
            yield client

    def _raise_for_status(self, response: httpx.Response):
        if response.status_code == 422:
            raise ValueError(response.json())
        elif response.status_code == 401:
            try:
                import streamlit
            except ImportError:
                response.raise_for_status()
            else:
                try:
                    response.raise_for_status()
                except Exception as exception:
                    if not self.streamlit_compatible:
                        streamlit_error_message = (
                            "\nStreamlit environment detected. Use StreamlitAuth.client instance.\n"
                            "Docs: https://developer.sweatstack.no/learn/integrations/streamlit/"
                        )
                        exception.add_note(streamlit_error_message)
                    raise

        else:
            response.raise_for_status()

    def _enums_to_strings(self, values: list[Enum | str]) -> list[str]:
        return [value.value if isinstance(value, Enum) else value for value in values]

    def _get_activities_generator(
        self,
        *,
        start: date | None = None,
        end: date | None = None,
        sports: list[Sport | str] | None = None,
        tags: list[str] | None = None,
        limit: int = 100,
    ) -> Generator[ActivitySummary, None, None]:
        num_returned = 0
        default_limit = 100
        params = {
            "limit": default_limit,
            "offset": 0,
        }
        if start is not None:
            params["start"] = start.isoformat()
        if end is not None:
            params["end"] = end.isoformat()
        if sports is not None:
            params["sports"] = self._enums_to_strings(sports)
        if tags is not None:
            params["tags"] = tags

        with self._http_client() as client:
            while True:
                response = client.get(
                    url="/api/v1/activities/",
                    params=params,
                )
                self._raise_for_status(response)
                activities = response.json()
                for activity in activities:
                    yield ActivitySummary.model_validate(activity)

                    num_returned += 1
                    if num_returned >= limit:
                        return
                if len(activities) < default_limit:
                    return

                params["limit"] = min(default_limit, limit - num_returned)
                params["offset"] += default_limit

    def _postprocess_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.streamlit_compatible:
            return make_dataframe_streamlit_compatible(df)
        else:
            return df

    def get_activities(
        self,
        *,
        start: date | None = None,
        end: date | None = None,
        sports: list[Sport | str] | None = None,
        tags: list[str] | None = None,
        limit: int = 100,
        as_dataframe: bool = False,
    ) -> list[ActivitySummary] | pd.DataFrame:
        """Gets a list of activities based on specified filters.

        Args:
            start: Optional start date to filter activities.
            end: Optional end date to filter activities.
            sports: Optional list of sports to filter activities by. Can be Sport objects or string IDs.
            tags: Optional list of tags to filter activities by.
            limit: Maximum number of activities to return. Defaults to 100.
            as_dataframe: Whether to return results as a pandas DataFrame. Defaults to False.

        Returns:
            Either a list of ActivitySummary objects or a pandas DataFrame containing
            the activities data, depending on the value of as_dataframe.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        activities = list(self._get_activities_generator(
            start=start,
            end=end,
            sports=sports,
            tags=tags,
            limit=limit,
        ))
        if as_dataframe:
            df = pd.DataFrame([activity.model_dump() for activity in activities])
            df = self._normalize_dataframe_column(df, "summary")
            df = self._normalize_dataframe_column(df, "laps")
            df = self._normalize_dataframe_column(df, "traces")
            return self._postprocess_dataframe(df)
        else:
            return activities

    def get_latest_activity(
        self,
        *,
        start: date | None = None,
        end: date | None = None,
        sport: Sport | None = None,
        tag: str | None = None,
    ) -> ActivityDetails:
        """Gets the most recent activity based on specified filters.

        Args:
            start: Optional start date to filter activities.
            end: Optional end date to filter activities.
            sport: Optional sport to filter activities by. Can be a Sport object or string ID.
            tag: Optional tag to filter activities by.

        Returns:
            ActivityDetails: The most recent activity matching the filters.

        Raises:
            StopIteration: If no activities match the filters.
            HTTPStatusError: If the API request fails.
        """
        return next(self._get_activities_generator(
            start=start,
            end=end,
            sports=[sport] if sport is not None else None,
            tags=[tag] if tag is not None else None,
            limit=1,
        ))

    def get_activity(self, activity_id: str) -> ActivityDetails:
        """Gets details for a specific activity by ID.

        Args:
            activity_id: The unique identifier of the activity to retrieve.

        Returns:
            ActivityDetails: The activity details object containing all information about the activity.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        with self._http_client() as client:
            response = client.get(url=f"/api/v1/activities/{activity_id}")
            self._raise_for_status(response)
            return ActivityDetails.model_validate(response.json())

    def get_activity_data(
        self,
        activity_id: str,
        adaptive_sampling_on: Literal["power", "speed"] | None = None,
    ) -> pd.DataFrame:
        """Gets the raw data for a specific activity.

        This method retrieves the time-series data for a given activity, with optional
        adaptive sampling to reduce data points for visualization.

        Args:
            activity_id: The unique identifier of the activity.
            adaptive_sampling_on: Optional parameter to apply adaptive sampling on 
                either "power" or "speed" data. If None, no adaptive sampling is applied.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the activity's time-series data.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        params = {}
        if adaptive_sampling_on is not None:
            params["adaptive_sampling_on"] = adaptive_sampling_on

        with self._http_client() as client:
            response = client.get(
                url=f"/api/v1/activities/{activity_id}/data",
                params=params,
            )
            self._raise_for_status(response)

        df = pd.read_parquet(BytesIO(response.content))
        return self._postprocess_dataframe(df)

    def get_activity_mean_max(
        self,
        activity_id: str,
        metric: Literal[Metric.power, Metric.speed] | Literal["power", "speed"],
        adaptive_sampling: bool = False,
    ) -> pd.DataFrame:
        """Gets the mean-max data for a specific activity.

        This method retrieves the mean-max curve data for a given activity, which represents
        the maximum average value of a metric (power or speed) for different time durations.

        Args:
            activity_id: The unique identifier of the activity.
            metric: The metric to calculate mean-max values for, either "power" or "speed".
            adaptive_sampling: Whether to apply adaptive sampling to reduce data points
                for visualization. Defaults to False.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the mean-max curve data.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        metric = self._enums_to_strings([metric])[0]
        with self._http_client() as client:
            response = client.get(
                url=f"/api/v1/activities/{activity_id}/mean-max",
                params={
                    "metric": metric,
                    "adaptive_sampling": adaptive_sampling,
                },
            )
            self._raise_for_status(response)
            df = pd.read_parquet(BytesIO(response.content))
            return self._postprocess_dataframe(df)

    def get_latest_activity_data(
        self,
        sport: Sport | str | None = None,
        adaptive_sampling_on: Literal["power", "speed"] | None = None,
    ) -> pd.DataFrame:
        """Gets the data for the latest activity of a specific sport.

        This method retrieves the time series data for the most recent activity of the specified sport.
        If no sport is specified, it returns data for the latest activity regardless of sport.

        Args:
            sport: Optional sport to filter by. Can be a Sport enum or string.
            adaptive_sampling_on: Optional metric to apply adaptive sampling for visualization.
                Can be either "power" or "speed". Defaults to None.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the activity data.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        activity = self.get_latest_activity(sport=sport)
        return self.get_activity_data(activity.id, adaptive_sampling_on)

    def get_latest_activity_mean_max(
        self,
        metric: Literal[Metric.power, Metric.speed] | Literal["power", "speed"],
        sport: Sport | str | None = None,
        adaptive_sampling: bool = False,
    ) -> pd.DataFrame:
        """Gets the mean-max curve for the latest activity of a specific sport.

        This method retrieves the mean-max curve data for the most recent activity of the specified sport.
        If no sport is specified, it returns data for the latest activity regardless of sport.

        Args:
            metric: The metric to calculate the mean-max curve for. Can be either "power" or "speed".
            sport: Optional sport to filter by. Can be a Sport enum or string.
            adaptive_sampling: Whether to apply adaptive sampling to the mean-max curve data.
                Defaults to False.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the mean-max curve data.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        activity = self.get_latest_activity(sport=sport)
        return self.get_activity_mean_max(activity.id, metric, adaptive_sampling)

    def get_longitudinal_data(
        self,
        *,
        sport: Sport | str | None = None,
        sports: list[Sport | str] | None = None,
        start: date | str,
        end: date | str | None = None,
        metrics: list[Metric | str] | None = None,
        adaptive_sampling_on: Literal[Metric.power, Metric.speed] | Literal["power", "speed"] | None = None,
    ) -> pd.DataFrame:
        """Gets longitudinal data for activities within a specified date range.

        This method retrieves aggregated data for activities that match the specified criteria,
        including sport type and date range. The data is returned as a pandas DataFrame.

        Args:
            sport: Optional single sport to filter by. Can be a Sport enum or string.
                Cannot be used together with 'sports'.
            sports: Optional list of sports to filter by. Can be a list of Sport enums or strings.
                Cannot be used together with 'sport'.
            start: The start date for the data range. Can be a date object or string in ISO format.
            end: Optional end date for the data range. Can be a date object or string in ISO format.
            metrics: Optional list of metrics to include in the results. Can be a list of Metric enums or strings.
            adaptive_sampling_on: Optional metric to apply adaptive sampling for visualization.
                Can be either "power" or "speed". Defaults to None.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the longitudinal activity data.

        Raises:
            ValueError: If both 'sport' and 'sports' parameters are provided.
            HTTPStatusError: If the API request fails.
        """
        if sport and sports:
            raise ValueError("Cannot specify both sport and sports")
        if sport is not None:
            sports = [sport]
        elif sports is None:
            sports = []

        params = {
            "sports": self._enums_to_strings(sports),
            "start": start
        }
        if end is not None:
            params["end"] = end
        if metrics is not None:
            params["metrics"] = self._enums_to_strings(metrics)
        if adaptive_sampling_on is not None:
            params["adaptive_sampling_on"] = self._enums_to_strings([adaptive_sampling_on])[0]

        with self._http_client() as client:
            response = client.get(
                url="/api/v1/activities/longitudinal-data",
                params=params,
            )
            self._raise_for_status(response)

            df = pd.read_parquet(BytesIO(response.content))
            return self._postprocess_dataframe(df)

    def get_longitudinal_mean_max(
        self,
        *,
        sport: Sport | str,
        metric: Literal[Metric.power, Metric.speed] | Literal["power", "speed"],
        date: date | str | None = None,
        window_days: int | None = None,
    ) -> pd.DataFrame:
        """Gets the mean-max curve for a specific sport and metric.

        This method retrieves the mean-max curve data for a given sport and metric,
        optionally filtered by date and window size.

        Args:
            sport: The sport to get mean-max data for. Can be a Sport enum or string ID.
            metric: The metric to calculate mean-max for. Must be either "power" or "speed".
            date: Optional reference date for the mean-max calculation. If provided,
                the mean-max curve will be calculated up to this date. Can be a date object
                or string in ISO format.
            window_days: Optional number of days to include in the calculation window
                before the reference date. If None, all available data is used.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the mean-max curve data.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        sport = self._enums_to_strings([sport])[0]
        metric = self._enums_to_strings([metric])[0]

        params = {
            "sport": sport,
            "metric": metric,
        }
        if date is not None:
            params["date"] = date
        if window_days is not None:
            params["window_days"] = window_days

        with self._http_client() as client:
            response = client.get(
                url="/api/v1/activities/longitudinal-mean-max",
                params=params,
            )
            self._raise_for_status(response)

            df = pd.read_parquet(BytesIO(response.content))
            return self._postprocess_dataframe(df)

    def _get_traces_generator(
        self,
        *,
        start: date | None = None,
        end: date | None = None,
        sports: list[Sport | str] | None = None,
        tags: list[str] | None = None,
        limit: int = 100,
    ) -> Generator[TraceDetails, None, None]:
        num_returned = 0
        default_limit = 100
        params = {
            "limit": default_limit,
            "offset": 0,
        }
        if start is not None:
            params["start"] = start.isoformat()
        if end is not None:
            params["end"] = end.isoformat()
        if sports is not None:
            params["sports"] = self._enums_to_strings(sports)
        if tags is not None:
            params["tags"] = tags

        with self._http_client() as client:
            while True:
                response = client.get(
                    url="/api/v1/traces/",
                    params=params,
                )
                self._raise_for_status(response)
                traces = response.json()
                for trace in traces:
                    yield TraceDetails.model_validate(trace)

                    num_returned += 1
                    if num_returned >= limit:
                        return
                if len(traces) < default_limit:
                    return

                params["limit"] = min(default_limit, limit - num_returned)
                params["offset"] += default_limit

    def _prepare_unserialized_data(self, df: pd.DataFrame, column: str) -> pd.DataFrame:
        """
        pd.json_normalize() only likes to play with lists of records (dicts?), not lists of lists.
        So that's what we're feeding it.
        """
        unserialized_data = df[column].tolist()
        if column in ["laps", "traces"]:
            result = []
            for sublist in unserialized_data:
                if sublist:
                    dict_from_sublist = {i: value for i, value in enumerate(sublist) if sublist}
                else:
                    dict_from_sublist = {}
                result.append(dict_from_sublist)

            unserialized_data = result

        return unserialized_data

    def _normalize_dataframe_column(self, df: pd.DataFrame, column: str) -> pd.DataFrame:
        normalized = pd.json_normalize(
            self._prepare_unserialized_data(df, column),
        )
        normalized = normalized.add_prefix(f"{column}.")
        normalized.index = df.index
        if column == "activity":
            normalized = normalized.drop(["activity.traces", "activity.laps"], axis=1, errors="ignore")
        return pd.concat([df.drop(column, axis=1), normalized], axis=1)

    def get_traces(
        self,
        *,
        start: date | None = None,
        end: date | None = None,
        sports: list[Sport | str] | None = None,
        tags: list[str] | None = None,
        limit: int = 100,
        as_dataframe: bool = False,
    ) -> list[TraceDetails] | pd.DataFrame:
        """Gets a list of traces based on specified filters.

        Args:
            start: Optional start date to filter traces.
            end: Optional end date to filter traces.
            sports: Optional list of sports to filter traces by. Can be Sport objects or string IDs.
            tags: Optional list of tags to filter traces by.
            limit: Maximum number of traces to return. Defaults to 100.
            as_dataframe: Whether to return results as a pandas DataFrame. Defaults to False.

        Returns:
            Either a list of TraceDetails objects or a pandas DataFrame containing
            the traces data, depending on the value of as_dataframe.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        traces = list(self._get_traces_generator(
            start=start,
            end=end,
            sports=sports,
            tags=tags,
            limit=limit,
        ))
        if not as_dataframe:
            return traces

        data = pd.DataFrame([trace.model_dump() for trace in traces])

        if "activity" in data.columns:
            data = self._normalize_dataframe_column(data, "activity")

        if "lap" in data.columns:
            data = self._normalize_dataframe_column(data, "lap")

        return self._postprocess_dataframe(data)

    def create_trace(
        self,
        *,
        timestamp: datetime,
        lactate: float | None = None,
        rpe: int | None = None,
        notes: str | None = None,
        power: int | None = None,
        speed: float | None = None,
        heart_rate: int | None = None,
        tags: list[str] | None = None,
    ) -> TraceDetails:
        """Creates a new trace with the specified parameters.

        This method creates a new trace entry with the given timestamp and optional
        measurement values.

        Args:
            timestamp: The date and time when the trace was recorded.
            lactate: Optional blood lactate concentration in mmol/L.
            rpe: Optional rating of perceived exertion (typically on a scale of 1-10).
            notes: Optional text notes associated with this trace.
            power: Optional power measurement in watts.
            speed: Optional speed measurement in meters per second.
            heart_rate: Optional heart rate measurement in beats per minute.
            tags: Optional list of tags to associate with this trace.

        Returns:
            TraceDetails: The created trace object with all details.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        with self._http_client() as client:
            response = client.post(
                url="/api/v1/traces/",
                json={
                    "timestamp": timestamp.isoformat(),
                    "lactate": lactate,
                    "rpe": rpe,
                    "notes": notes,
                    "power": power,
                    "speed": speed,
                    "heart_rate": heart_rate,
                    "tags": tags,
                },
            )
            self._raise_for_status(response)
            return TraceDetails.model_validate(response.json())

    def get_sports(self, only_root: bool = False) -> list[Sport]:
        """Gets a list of available sports.

        This method retrieves all sports available to the user, with an option to only
        return root sports (top-level sports without parents).

        Args:
            only_root: If True, only returns root sports without parents. Defaults to False.

        Returns:
            list[Sport]: A list of Sport objects representing the available sports.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        with self._http_client() as client:
            response = client.get(
                url="/api/v1/profile/sports/",
                params={"only_root": only_root},
            )
            self._raise_for_status(response)
            return [Sport(sport) for sport in response.json()]

    def get_tags(self) -> list[str]:
        """Gets a list of all tags used by the user.

        This method retrieves all tags that the user has created or used across
        their activities and traces.

        Returns:
            list[str]: A list of tag strings.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        with self._http_client() as client:
            response = client.get(
                url="/api/v1/profile/tags/",
            )
            self._raise_for_status(response)
            return response.json()

    def get_users(self) -> list[UserSummary]:
        """Gets a list of all users accessible to the current user.

        This method retrieves all users that the current user has access to view.
        For regular users, this typically returns only their own user information.
        For admin users, this may return information about multiple users.

        Returns:
            list[UserSummary]: A list of UserSummary objects containing basic user information.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        with self._http_client() as client:
            response = client.get(
                url="/api/v1/users/",
            )
            self._raise_for_status(response)
            return [UserSummary.model_validate(user) for user in response.json()]

    def get_userinfo(self) -> UserInfoResponse:
        """Gets detailed information about the current user.

        This method retrieves comprehensive information about the user currently
        authenticated with the client.

        Returns:
            UserInfoResponse: A UserInfoResponse object containing detailed user information
                including profile data, permissions, and authentication details.

        Raises:
            HTTPStatusError: If the API request fails.
        """
        with self._http_client() as client:
            response = client.get(
                url="/api/v1/oauth/userinfo",
            )
            self._raise_for_status(response)
            return UserInfoResponse.model_validate(response.json())


_default_client = Client()


def _generate_singleton_methods(method_names: List[str]) -> None:
    """
    Automatically generates singleton methods for the Client class.
    
    Args:
        method_names: List of method names to expose in the singleton interface
    """

    def create_singleton_method(method_name: str):
        bound_method = getattr(_default_client, method_name)

        @wraps(bound_method)
        def singleton_method(*args: Any, **kwargs: Any) -> Any:
            return bound_method(*args, **kwargs)

        class_method = getattr(Client, method_name)
        singleton_method.__annotations__ = get_type_hints(class_method)

        return singleton_method
    
    for method_name in method_names:
        if not hasattr(Client, method_name):
            raise ValueError(f"Method '{method_name}' not found in class {Client.__name__}")
            
        class_method = getattr(Client, method_name)
        
        if not callable(class_method):
            continue
            
        globals()[method_name] = create_singleton_method(method_name)


_generate_singleton_methods(
    [
        "login",

        "get_users",
        "get_userinfo",

        "get_activities",

        "get_activity",
        "get_activity_data",
        "get_activity_mean_max",

        "get_latest_activity",
        "get_latest_activity_data",
        "get_latest_activity_mean_max",

        "get_longitudinal_data",
        "get_longitudinal_mean_max",

        "get_traces",
        "create_trace",

        "get_sports",
        "get_tags",

        "switch_user",
        "switch_back",
        "delegated_client",
        "principal_client",
    ]
)